/*
  setuidgid.c
  Matlib 2010-02-24
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>

#include <arguments.h>


#define CHECK_ERROR(__COND__, __FUN__) \
  if (__COND__) \
  { \
    perror (__FUN__); \
    exit (EXIT_FAILURE); \
  }


struct passwd *defaults;
uid_t get_uid (char *name)
{
  char *tail;
  long int uid;

  uid = strtol (name, &tail, 0);
  if ((*tail == 0) && (uid >= 0))
    return uid;
  if ((defaults = getpwnam (name)) == NULL)
    arguments_error_freeform ("User \"%s\" not found in password database.", 
       name);
  return defaults -> pw_uid;
}

gid_t get_gid (char *name)
{
  char *tail;
  long int gid;
  struct group *g;

  gid = strtol (name, &tail, 0);
  if ((*tail == 0) && (gid >= 0))
    return gid;
  if ((g = getgrnam (name)) == NULL)
    arguments_error_freeform ("Group \"%s\" not found in group database.", 
       name);
  return g -> gr_gid;
}

void get_defaults ()
{
  if (defaults != NULL)
    return;
  defaults = getpwuid (getuid ());
  if (defaults == NULL)
  {
    fprintf (stderr, "Cannot retrieve user settings for UID %d.", getuid ());
    exit (EXIT_FAILURE);
  }
}




ARGUMENT_CALLBACK (argument_uid)
{
  arg -> value = (void *) (long int) get_uid (parameter);
}

ARGUMENT_CALLBACK (argument_gid)
{
  arg -> value = (void *) (long int) get_gid (parameter);
}

gid_t *groups;
int gp, gl;
#define GROUP_BLOCK 4
ARGUMENT_CALLBACK (argument_groups)
{
  if (gp == gl)
  {
    gl += GROUP_BLOCK;
    groups = realloc (groups, sizeof (gid_t) * gl);
    CHECK_ERROR (groups == NULL, "realloc");
  }
  groups [gp] = get_gid (parameter);
  gp ++;
}

struct Argument arguments [] =
{
  {'u', "ruser",        argument_uid,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) -1},
  {'U', "euser",        argument_uid,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) -1},
  {'g', "rgroup",       argument_gid,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) -1},
  {'G', "egroup",       argument_gid,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) -1},
  {'c', "clear-groups", argument_bool,   ARGUMENT_F_ONCE, NULL},
  {'s', "add-group",    argument_groups, ARGUMENT_F_PAR, NULL},
  {'n', "program-name", argument_string, ARGUMENT_F_ONCE | ARGUMENT_F_PAR, NULL},
  {}
};
#define ARG_IDX_RUID 0
#define ARG_IDX_EUID 1
#define ARG_IDX_RGID 2
#define ARG_IDX_EGID 3
#define ARG_IDX_CLRG 4
#define ARG_IDX_ADDG 5
#define ARG_IDX_NAME 6
#define ARG_RUID ((long int) (arguments [ARG_IDX_RUID].value))
#define ARG_EUID ((long int) (arguments [ARG_IDX_EUID].value))
#define ARG_RGID ((long int) (arguments [ARG_IDX_RGID].value))
#define ARG_EGID ((long int) (arguments [ARG_IDX_EGID].value))
#define ARG_CLRG (arguments [ARG_IDX_CLRG].value != NULL)
#define ARG_NAME ((char *) (arguments [ARG_IDX_NAME].value))




void set_defaults (char *name)
{
  arguments [ARG_IDX_RUID].value = arguments [ARG_IDX_EUID].value = 
     (void *) (long int) get_uid (name);
  get_defaults ();
  arguments [ARG_IDX_RGID].value = arguments [ARG_IDX_EGID].value =
     (void *) (long int) defaults -> pw_gid;
  arguments [ARG_IDX_CLRG].value = (void *) 1;
}


void set_groups ()
{
  gid_t *g;
  int n, i, j;
  
  if (ARG_CLRG)
  {
    CHECK_ERROR (setgroups (gp, groups) != 0, "setgroups");
    return;
  }
  n = getgroups (0, NULL);
  CHECK_ERROR (n == -1, "getgroups: get number of groups");
  if (n == 0)
  {
    CHECK_ERROR (setgroups (gp, groups) != 0, "setgroups");
    return;
  }
  g = malloc (sizeof (gid_t) * (n + gp));
  i = getgroups (n, g);
  CHECK_ERROR (i == -1, "getgroups");
  if (i != n)
  {
    fprintf (stderr, "Calls to getgroups returned different number of "
       "supplementary groups. Confused, aborting.\n");
    exit (EXIT_FAILURE);
  }
  
  i = 0;
  while (i < gp)
  {
    for (j = 0; j < n; j ++)
      if (g [j] == groups [i])
        goto lbreak;
    g [n ++] = groups [i];
lbreak:
    i ++;
  }
  CHECK_ERROR (setgroups (n, g) != 0, "setgroups");
}


int main (int argc, char **argv)
{
  int n;
  char **a;
  char *progname;
  
  defaults = NULL;
  gp = 0;
  gl = GROUP_BLOCK;
  groups = malloc (sizeof (gid_t) * GROUP_BLOCK);
  CHECK_ERROR (groups == NULL, "malloc");
  n = arguments_parse ("setuidgid", "1 (2010-02-24)", "[command ...]", 
     "Set real or effective process IDs and execute a command.",
     argc, argv, arguments);
     
  if ((ARG_RUID == -1) && (ARG_RGID == -1) && (ARG_EUID == -1)
     && (ARG_RGID == -1))
  {
    if (n == argc)
      arguments_error_freeform ("User name required.");
    set_defaults (argv [n]);
    n ++;
  }
  
  if (ARG_CLRG || (gp != 0))
    set_groups ();
  CHECK_ERROR (setregid (ARG_RGID, ARG_EGID) != 0, "setregid");
  CHECK_ERROR (setreuid (ARG_RUID, ARG_EUID) != 0, "setreuid");
  
  if (n == argc)
  {
    get_defaults ();
    progname = defaults -> pw_shell;
    if (progname == NULL)
    {
      fprintf (stderr, 
         "User's default shell is not specified in password database.\n");
      exit (EXIT_FAILURE);
    }
  }
  else
    progname = argv [n ++];
  a = malloc (sizeof (char *) * (argc - n + 2));
  a [0] = ARG_NAME == NULL ? progname : ARG_NAME;
  for (gp = 1; n < argc; gp ++)
    a [gp] = argv [n ++];
  a [gp] = NULL;
  execvp (progname, a);
  perror (progname);
  return EXIT_FAILURE;
}

